/*
  Copyright (C) 2002-2003 Laurent Martelli <laurent@aopsys.com>
                          Renaud Pawlak <renaud@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.aspects.gui;

import dori.jasper.engine.JRException;
import dori.jasper.engine.JasperCompileManager;
import dori.jasper.engine.JasperExportManager;
import dori.jasper.engine.JasperFillManager;
import dori.jasper.engine.JasperPrint;
import dori.jasper.engine.JasperReport;
import org.objectweb.jac.aspects.authentication.AuthenticationAC;
import org.objectweb.jac.aspects.gui.DisplayContext;
import org.objectweb.jac.aspects.gui.reports.JacDataSource;
import org.objectweb.jac.aspects.session.SessionAC;
import org.objectweb.jac.core.ACManager;
import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.Display;
import org.objectweb.jac.core.NameRepository;
import org.objectweb.jac.core.rtti.AbstractMethodItem;
import org.objectweb.jac.core.rtti.ClassRepository;
import org.objectweb.jac.util.Log;
import java.io.File;
import java.util.HashMap;

/**
 * This class is a container of commonly used GUI actions static
 * methods.
 *
 * <p>Usually, a GUI action takes in parameter the display
 * context. Using the context, is can then know the customized window
 * to use to interact with the users. When an action does not take any
 * display context, it means that it does not need any interaction
 * with the GUI or that the display context is retrieved throught the
 * collaboration. */

public class Actions {

   /**
    * Quit the application by shutting down the server.
    *
    * @param context the display context for this action
    */
   public static void exit(DisplayContext context) {
      Display display = context.getDisplay();
      if (display!=null)
         display.close();
      System.exit(0);
   }
   
   /**
    * Logout from an application by clearing the current user. If an
    * authentication aspect is woven, the user is asked again.
    *
    * @param context the display context for this action */

   public static void logout(DisplayContext context) {
      SessionAC sac = ((SessionAC)ACManager.getACM().getAC("session"));
      if (sac==null) {
         Log.error("No session aspect found: logout cannot work");
         return;
      }
      sac.clearCurrentSessionAttribute(AuthenticationAC.USER);
      Collaboration.get().removeAttribute(SessionAC.INITIALIZED);
      context.getDisplay().fullRefresh();
   }

   /**
    * Shows a message on the status bar of the current customized. If
    * no display context available, then <code>println</code> is
    * used.
    * 
    * @param message the message to show */

   public static void showStatus(String message) {
      CustomizedView cview = ((DisplayContext)Collaboration.get()
         .getAttribute(GuiAC.DISPLAY_CONTEXT)).getCustomizedView();
      if(cview!=null) {
         cview.showStatus(message);
      }
   }

   /**
    * Show a customized window that has been declared and configured
    * within the GUI aspect.
    *
    * @param context the display context for this action
    * @param id the customized ID */

   public static void showWindow(DisplayContext context,String id) {
      ((GuiAC)ACManager.getACM().getAC("gui"))
         .createSwingDisplays(new String[] {id});
   }
  
   /**
    * Invokes a method on an object. */

   public static void invoke(Object object, String methodName) {
      AbstractMethodItem method = ClassRepository.get()
         .getClass(object).getMethod(methodName);
      EventHandler.get().onInvoke((DisplayContext)Collaboration.get()
         .getAttribute(GuiAC.DISPLAY_CONTEXT),object,method);
   }

   /**
    * Display an object in a panel.
    *
    * @param context the DisplayContext of the custmoized view
    * @param objectName the name of the object to display
    * @param panelID the panel ID where to display the object
    */
   public static void viewObject(DisplayContext context, 
                                 String objectName, String panelID) {
      CustomizedView custom = context.getCustomizedView();
      if (custom!=null) {
         ViewFactory factory = context.getDisplay().getFactory();
         Object object = NameRepository.get().getObject(objectName);
         if (object!=null)
            custom.getPanelView().addView(
               factory.createView(
                  objectName,"Object",
                  new Object[] {GuiAC.DEFAULT_VIEW,object},
                  context),
               panelID
            );
         else 
            custom.getPanelView().addView(
               factory.createView("No such object "+objectName,"Label",context),
               panelID
            );

         EventHandler.get().maybeInvalidatePane(
            context.getDisplay().getFactory(),context,
            custom,panelID);
      }
   }

   /**
    * Generate a PDF report
    * @param reportDef resource name of the XML report definition file
    * @param pdfFile file where to store the resulting PDF document
    */
   public static void genReport(String reportDef, File pdfFile) 
      throws JRException 
   {
      Log.trace("report","Compiling report file "+reportDef);
      JasperReport report = 
         JasperCompileManager.compileReport(
            Actions.class.getClassLoader().getResourceAsStream(reportDef));
      Log.trace("report","Filling report "+report);
      JasperPrint print = 
         JasperFillManager.fillReport(
            report,
            new HashMap(),
            new JacDataSource(ClassRepository.get().getClass(Class.class)));
      Log.trace("report","Exporting report "+print);
      JasperExportManager.exportReportToPdfFile(print,pdfFile.getPath());
      Log.trace("report","Done");
   }

   /**
    * Changes a trace for the current application
    * @param category category of the trace
    * @param level level of the trace (0=quiet,1=verbose,2=more verbose...)
    */
   public static void setTrace(String category, int level) {
      Log.setLevel(category,level);      
   }

   /**
    * Reloads an aspect for the current application
    */
   public static void reloadAspect(String aspect) throws Exception {
      ACManager.getACM().reloadAspect(aspect);
   }

   /**
    * Enable EJP profiling
    */
   /*
   public static void enableProfiling() throws Throwable {
      if (!ejp.tracer.TracerAPI.enableTracing()) {
         throw ejp.tracer.TracerAPI.getInitializationError();
      }
   }

   public static void disableProfiling() throws Throwable {
      if (!ejp.tracer.TracerAPI.disableTracing()) {
         throw ejp.tracer.TracerAPI.getInitializationError();
      }
   }
   */
}
