/*
  Copyright (C) 2001-2003 Renaud Pawlak <renaud@aopsys.com>, 
                          Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.gui;

import org.objectweb.jac.core.rtti.*;

/**
 * This configuration interface of the Gui aspect defines all the
 * methods that handle the appearence of the fields in objects views.
 *
 * @see GuiAC
 * @see View
 *
 * @author <a href="mailto:renaud@cnam.fr">Renaud Pawlak</a>
 * @author <a href="mailto:laurent@aopsys.com">Laurent Martelli</a> */

public interface FieldAppearenceGuiConf {

   /**
    * Sets an item to be visible or not.
    *
    * <p>If a field of method item is not visible, it will not be
    * displayed by the object views. If a class item is not visible,
    * then it will not be possible to open a view on objects of this
    * class.
    *
    * <p>By default, the RTTI does not define this property and every
    * item is visble. Thus, the GUI aspect component programmer should
    * overload the default contructor of this class to set some
    * visible attributes to false.<p>
    * 
    * @param member the member item (may be a method or a field name);
    * it can be null to set the whole class visible or not
    * @param visible whether the member must be visible
    *
    * @see GuiAC#isVisible(MetaItem) */

   void setVisible(MemberItem member, boolean visible);

   /**
    * Tells the GUI to use embedded editors for all fields of a class,
    * whenever possible.
    * @param cl the class item
    * @see FieldEditor */

   void setEmbeddedEditors(ClassItem cl);

   /**
    * This configuration method allows the programmer to set a class
    * member to be internally edited.
    *
    * <p>By default, each member of a class is editable with an "edit"
    * button or link that opens a popup that allows the user to edit
    * its value (see the <code>FieldEditor</code> interface). If this
    * configuration method is called, then the editor will be embedded
    * into the object's view the member belongs to.
    *
    * @param member the member item to be embedded (may be a
    * field or a method)
    * @see FieldEditor */

   void setEmbeddedEditor(MemberItem cl);

   /**
    * Sets a border to the field in an object view.
    *
    * @param field the field
    * @param alignment LEFT: title is on the left, RIGHT:
    * title is on the right, CENTER: title is centered 
    * @param style LINE: the border is a line, ETCHED: the border is a
    * 3D line, LOWERED: the border is a 3D effect that makes the
    * bordered element lowered, RAISED: the border is a 3D effect that
    * makes the bordered element raised.  
    * @see #setBorder(FieldItem)
    */
   void setBorder(FieldItem field,String alignment,String style);

   /**
    * Sets a border to the field in an object view. Uses a default
    * style (LINE) and alignement(LEFT)
    *
    * @param field the field
    * @see #setBorder(FieldItem,String,String)
    */
   void setBorder(FieldItem field);

   /**
    * Sets the width of the field's embedded editor when exist.
    *
    * @param field the field
    * @param width the editor width in pixel (typically used by
    * SWING)
    * @param columns the editor width in character columns (typically
    * used by the WEB)
    * @see #setEmbeddedEditor(MemberItem) 
    * @see #setEditorHeight(FieldItem,int,int)
    * @see #setDefaultEditorWidth(ClassItem,int,int)
    */
   void setEditorWidth(FieldItem field, int width, int columns);

   /**
    * Sets the default editor width for value of a given type.
    *
    * @param field the field 
    * @param width the editor width in pixel (typically used by
    * SWING)
    * @param columns the editor width in character columns (typically
    * used by the WEB)
    *
    * @see #setDefaultEditorWidth(ClassItem,int,int)
    * @see #setDefaultEditorHeight(VirtualClassItem,int,int)
    * @see #setEditorWidth(FieldItem,int,int)
    */
   void setDefaultEditorWidth(VirtualClassItem type, int width, int columns);

   /**
    * Sets the default editor width for value of a given type.
    *
    * @param field the field 
    * @param width the editor width in pixel (typically used by
    * SWING)
    * @param columns the editor width in character columns (typically
    * used by the WEB)
    *
    * @see #setDefaultEditorWidth(VirtualClassItem,int,int)
    * @see #setDefaultEditorHeight(VirtualClassItem,int,int)
    * @see #setEditorWidth(FieldItem,int,int)
    */
   void setDefaultEditorWidth(ClassItem type, int width, int columns);

   /**
    * Set a field to be editable.
    *
    * @param field the field
    * @param editable the flag (true is default) */

   void setEditable(FieldItem field,boolean editable);

   /**
    * Sets the height of the field's embedded editor when exist.
    *
    * @param field the field 
    * @param height the editor height in pixel
    * @param rows the editor height in character rows
    *
    * @see #setEmbeddedEditor(MemberItem) 
    * @see #setEditorWidth(FieldItem,int,int)
    * @see #setDefaultEditorHeight(ClassItem,int,int)
    */
   void setEditorHeight(FieldItem field, int height, int rows);

   /**
    * Sets the default editor height for value of a given type.
    *
    * @param field the field 
    * @param height the editor height in pixel
    * @param rows the editor height in character rows
    *
    * @see #setDefaultEditorHeight(ClassItem,int,int)
    * @see #setDefaultEditorWidth(VirtualClassItem,int,int)
    * @see #setEditorHeight(FieldItem,int,int)
    */
   void setDefaultEditorHeight(VirtualClassItem type, int height, int rows);

   /**
    * Sets the default editor height for value of a given type.
    *
    * @param field the field 
    * @param height the editor height in pixel
    * @param rows the editor height in character rows
    *
    * @see #setDefaultEditorHeight(VirtualClassItem,int,int)
    * @see #setDefaultEditorWidth(VirtualClassItem,int,int)
    * @see #setEditorHeight(FieldItem,int,int)
    */
   void setDefaultEditorHeight(ClassItem type, int height, int rows);

   /**
    * Sets the category of an item of a class.
    *
    * <p>The category must correspond to one of these defined on the
    * class with the <code>setCategories</code> configuration method.
    * 
    * @param member the member to categorize (a method or a field)
    * @param category the existing category name
    * @see #setCategories(MemberItem,String[]) 
    * @see ClassAppearenceGuiConf#setCategories(ClassItem,String[]) 
    */
   void setCategory(MemberItem member, String category);

   /**
    * Sets the categories of an item of a class.
    *
    * <p>The category must correspond to one of these defined on the
    * class with the <code>setCategories</code> configuration method.
    * 
    * @param member the member to categorize (a method or a field)
    * @param categories the existing category names
    * @see #setCategory(MemberItem,String)
    * @see ClassAppearenceGuiConf#setCategories(ClassItem,String[]) 
    */
   void setCategories(MemberItem member, String[] categories);

   /**
    * Tells the GUI to insert a referenced object to be displayed as an
    * embedded view in its container object view.
    *
    * @param member the member (reference field or method) that must
    * be embedded */
   void setEmbeddedView(MemberItem member);

   /**
    * Tells the GUI wether to use an embedded view for the adder of a
    * collection.  
    * @param collection the collection
    * @param embedded wether to use an embedded view for the adder
    */
   void setEmbeddedAdder(CollectionItem collection, boolean embedded);

   /**
    * Sets the render of a given field (more precisely a collection)
    * to be rendered by a table.
    *
    * <p>In a table view, each item of the displayed collection fills
    * a table line. Each column represents one field of the objects
    * whithin the collection (the item within the collection whould be
    * of the same class --- or at least share a common superclass).
    *
    * @param field the field that contains the collection 
    */
   void setTableView(FieldItem field);

   /**
    * Tells the GUI wether to show row numbers for tables and lists.
    * @param collection the affected collection
    * @param value wether to show row numbers
    *
    * @see setDefaultShowRowNumbers(boolean)
    */
   void showRowNumbers(CollectionItem collection, boolean value);

   /**
    * Tells the GUI wether to show row numbers for tables and lists by
    * default.
    *
    * @param value wether to show row numbers
    *
    * @see showRowNumbers(CollectionItem,boolean) 
    */
   void setDefaultShowRowNumbers(boolean value);

   /**
    * Sets a default sorted column for a collection.
    *
    * <p>By default, collections are not sorted. You can precise a
    * column to use to sort the collection by default. It will be used
    * at the construction of the collection.</p>
    *
    * @param collection the collection
    * @param column the column used to sort (it is a fieldItem, watch
    * out for case). You may preprend a '-' to use the reverse order
    * of that column.
    */
   void setDefaultSortedColumn(CollectionItem collection, String column);

   /**
    * Sets the default order in which the attributes of the elements
    * of a collection are to be rendered.
    *
    * @param collection the collection
    * @param targetClass the class of attributes to render
    * @param memberNames the name of the members in the
    * rendering order
    *
    * @see #setMembersOrder(CollectionItem,String,ClassItem,String[])
    * @see ClassAppearenceGuiConf#setTableMembersOrder(ClassItem,String[])
    * @see ClassAppearenceGuiConf#setAttributesOrder(ClassItem,String[]) */
   void setMembersOrder(CollectionItem collection, 
                        ClassItem targetClass, String[] memberNames);

   /**
    * Sets the order in which the attributes of the elements of a
    * collection are to be rendered for a given view.
    *
    * @param collection the collection
    * @param viewName the type for which to set the members order
    * @param targetClass the class of attributes to render
    * @param memberNames the name of the members in the
    * rendering order
    * 
    * @see #setMembersOrder(CollectionItem,ClassItem,String[]) */
   void setMembersOrder(CollectionItem collection, String viewName,
                        ClassItem targetClass, String[] memberNames);

   /**
    * This configuration method allows the programmer to make a set of
    * object to be proposed to the final user when an edition of this
    * field value is performed.
    *
    * <p>Most of the GUI will propose the choice within a ComboBox.
    *
    * <p>When the choices values can not be defined at programming
    * time but must be dynamically created, then the programmer can
    * use the <code>setDynamicFieldChoice</code> method.
    *
    * @param field the field
    * @param choice the values the user will have to choose from when
    * a edition of the field is performed
    *
    * @see #setDynamicFieldChoice(FieldItem,Boolean,ClassItem,String) 
    * @see #setFieldChoice(FieldItem,Boolean,String[])
    * @see #setFieldEnum(FieldItem,String)
    * @see GuiConf#defineEnum(String,String[],int,int)
    */
   void setFieldChoice(FieldItem field, 
                       Boolean editable, String[] choice);


   /**
    * <p>Declare a field as an enumeration.</p>
    *
    * @param field the field
    * @param enum the name of the enumeration
    *
    * @see GuiConf#defineEnum(String,String[],int,int)
    * @see #setFieldChoice(FieldItem,Boolean,String[])
    * @see #setDynamicFieldChoice(FieldItem,Boolean,ClassItem,String) 
    */
   void setFieldEnum(FieldItem field, String enum);

   /**
    * Same as <code>setFieldChoice</code> but with dynamically defined values.
    *
    * <p>The values are dynamically defined at runtime by the
    * invocation of a target method. This target method must return a
    * collection of objects or an array of objects that contains the
    * possible new values for the fields.</p>
    *
    * <p>If the target method is static, it will be called with the
    * object as the only parameter. If it's not static, it will called
    * <em>on</em> the object with no parameters.</p>
    *
    * @param field the field
    * @param targetClass the class that contains the target method
    * @param targetMethod name of a static method within the target
    * class that returns the values to choose from. It must take an
    * Object as parameter which will be the instance to which the
    * field belongs to. */
   void setDynamicFieldChoice(FieldItem field, Boolean editable,
                              ClassItem targetClass, String targetMethod);

   /**
    * Same as <code>setFieldChoice</code> but with dynamically defined values.
    * @param field the field
    * @param targetMethod a static method that returns the values to
    * choose from. It must take an Object as parameter which will be
    * the instance to which the field belongs to.
    * @see #setDynamicFieldChoice(FieldItem,Boolean,ClassItem,String) */
   void setDynamicFieldChoice(FieldItem field, Boolean editable,
                              MethodItem targetMethod);

   /**
    * Set the type of the objects of a collection.
    *
    * <p>If this method is not used, the collection type can be
    * dynamically found out by the GUI from the adder's argument
    * types.
    *
    * @param collection the collection within this class 
    * @param type the type of this collection (an exiting class name) 
    */
   void setCollectionType(CollectionItem collection, String type);

   /**
    * Tells the preferred height a table or list view of a collection
    * should take, if possible.
    *
    * @param collection the collection
    * @param height the preferred height 
    */
   void setPreferredHeight(CollectionItem collection, int height);

   /**
    * Tells the preferred width a table or list view of a collection
    * should take, if possible.
    *
    * @param collection the collection
    * @param width the preferred width 
    */
   void setPreferredWidth(CollectionItem collection, int width);

   /**
    * Sets the number of rows to display simultaneously for a
    * collection. This is only used by the web GUI so that generated
    * web pages are not too big. If the number of elements in the
    * collection is bigger than numRows, a "previous" and a "next"
    * button are displayed so that the user can see the rest of the
    * collection. The default is 10.
    *
    * @param collection the collection
    * @param numRows the number of rows per page 
    */
   void setNumRowsPerPage(CollectionItem collection, int numRows);

   /**
    * Sets the view of a given setter's calling box to be a file
    * chooser.<p>
    *
    * As logically expected, the type of the set field must be a
    * string or an URL.<p>
    *
    * @param method the method item. It can be of the form
    * "methodName" or "methodName(<types>)". The first syntax will use
    * the method with that name. The second syntax allow you to
    * specify parameter types (separated by commas, with no spaces).
    * @param paramNames the parameter names
    * @see java.net.URL
    * @see GuiAC#isFileChooserView(MethodItem) */

   void setFileChooserView(MethodItem method, 
                           String[] fileExtensions,
                           String fileDescription );

   /**
    * This configuration method attaches an icon to a given field so
    * that the iconized instances of this field will be represented by
    * this icon (for instance in a treeview).
    *
    * @param member the member (field or method)
    * @param name the icon's resource name 
    */
   void setIcon(MemberItem field, String name);

   /**
    * This configuration method allows to hide a given relation
    * (collection) in a treeview, even if the show relations mode is
    * on .
    *
    * @param field the field
    */   
   void hideTreeRelation(FieldItem field);

   /**
    * Set the default value for a choice.
    *
    * <p>The default value is dynamically calculated by the
    * <code>method</code> parameter from the string value (result can
    * be an object). A default implementation for method is provided
    * by <code>GuiAC</code>.
    *
    * @param field the field
    * @param method a static method that returns the default value
    * (prototype: Object m(FieldItem,String))
    * @param value the string representation of the default value
    * @see GuiAC#getDefaultValue(FieldItem,String) */
   void setDefaultValue(FieldItem field,
                        MethodItem method, String value);

   /**
    * Set the default value for a choice.
    *
    * <p>Same as setDefaultValue(ClassItem, String, MethodItem, String)
    * using default MethodItem <code>GuiAC.getDefaultValue</code>.
    *
    * @param field the field
    * @param value the string representation of the default value
    * @see GuiAC#getDefaultValue(FieldItem,String)
    * @see #setDefaultValue(FieldItem,MethodItem,String) 
    */
   void setDefaultValue(FieldItem field, String value);

   /**
    * Sets the display format of a float or double field.
    *
    * @param field the field
    * @param format the display format of the field
    *
    * @see java.text.DecimalFormat
    */
   void setFloatFormat(FieldItem field, String format);

   /**
    * Sets the description of a class member (field of method).
    *
    * @param member the member
    * @param description the description of the class member
    */
   void setDescription(MemberItem member, String description);


   /**
    * Sets the label of a class member (field of method).
    *
    * @param member the member
    * @param label the label of the class member
    */
   void setLabel(MemberItem member, String label);

   /**
    * Wether to display a label containing the name of the field in views.
    *
    * @param member the member
    * @param value boolean indicating wether to display the label
    */
   void setDisplayLabel(MemberItem member, boolean value);


   /**
    * Set the style of a field
    *
    * @param field the field
    * @param style the CSS style
    *
    * @see ClassAppearenceGuiConf#setStyle(ClassItem,String)
    */
   void setStyle(FieldItem field, String style);

   /**
    * Sets the view type of a field instead of using the default
    * one ("List" or "Table"). It allows you to define a custom view
    * constructor for that type in order to handle complex tables.
    * The view constructor must take 3 arguments: a CollectionItem, an
    * Object (the susbtance), and a CollectionItemView)
    */
   void setViewType(FieldItem field, String viewName,
                    String viewType);

}
