/*
  Copyright (C) 2002-2003 Renaud Pawlak <renaud@aopsys.com>,
                          Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.aspects.gui;

import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.rtti.ClassRepository;
import org.objectweb.jac.core.rtti.CollectionItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.util.Log;
import javax.swing.tree.TreePath;
import org.objectweb.jac.util.Stack;

/**
 * This class represents a tree node for an object. */

public class ObjectNode extends AbstractNode 
   implements ObjectUpdate, FieldUpdate, CollectionUpdate 
{
   FieldItem relation;
   Object substance;

   /**
    * Constructor.
    *
    * @param model the tree modelto notify when changes occur
    * @param value the value that the node represents
    * @param substance the owner of the relation
    * @param relation the relation the value is part of
    */
   public ObjectNode(TreeView model, Object value, 
                     Object substance, FieldItem relation) {
      super(model,value);
      this.substance = substance;
      this.relation = relation;
      Utils.registerObject(value,this);
      if (!(relation instanceof CollectionItem)) {
         Utils.registerField(value,relation,this);
      }
      if (GuiAC.getGraphicContext()!=null)
         context.addAll(GuiAC.getGraphicContext());
      if (relation!=null)
         context.push(relation);
      rebuildData();
   }

   Stack context = new Stack();

   /**
    * Returns the relation (reference or collection) the substance of
    * the node belongs to. */
   public FieldItem getRelation() {
      return relation;
   }

   /**
    * Returns the substance of this node. */
   public Object getSubstance() {
      return substance;
   }

   /**
    * Rebuild the data of this node again. */

   protected void rebuildData() {
      Object object = getUserObject();
      Log.trace("gui.treeview","refresh("+object+")");
      if (object==null) {
         icon = null;
         text = null;
      } else {
         text = GuiAC.toString(object,context);
         Log.trace("gui.treeview",2,"text="+text);
         icon = GuiAC.getIcon(ClassRepository.get().getClass(object));
      }
   }

   /**
    * Unregisters from the events this node is notified. */

   public void unregisterEvents() {
      Object value = getUserObject();
      Utils.unregister(value,this);
   }

   /**
    * Find a node in the children.
    * @param relation the relation the requested node must be part of
    * @param userObject the userObject value the requested node must have
    * @return an ObjectNode with the requested features, or null if no
    * such node can be found.
    */
   protected ObjectNode findNode(FieldItem relation, Object userObject) {
      Log.trace("gui.treeView","Looking for node "+relation+" -> "+userObject);
      for (int i=0; i<getChildCount(); i++) {
         if (getChildAt(i) instanceof ObjectNode) {
            ObjectNode current = (ObjectNode)getChildAt(i);
            if (current.getUserObject()==userObject &&
                current.getRelation()==relation)
               return current;
         }
      }
      return null;
   }

   // ObjectUpdate interface

   public void objectUpdated(Object substance, Object param) {
      rebuildData();
      model.nodeChanged(this);
   }

   // FieldUpdate interface

   public void fieldUpdated(Object substance, FieldItem collection, 
                            Object value, Object param) {
      rebuildData();
      model.nodeChanged(this);
   }

   // CollectionUpdate interface

   public void onChange(Object substance, CollectionItem collection, 
                        Object value, Object param) {
      Log.trace("gui.collectionUpdate",
                "ObjectNode collectionUpdated "+collection.getLongName());
      // remove all children nodes
      int[] indices = new int[getChildCount()];
      for(int i=0; i<indices.length; i++) {
         indices[i] = i;
      }
      AbstractNode[] removedNodes = new AbstractNode[getChildCount()];
      for(int i=0;i<indices.length;i++) {
         removedNodes[i] = (AbstractNode)getChildAt(i);
      }
      removeAllChildren();
      model.nodesWereRemoved(this,indices,removedNodes);

      // rebuild children nodes
      TreeModel.addNodes(model,this,getUserObject(),showRelations);

      indices = new int[getChildCount()];
      for(int i=0;i<indices.length;i++) {
         indices[i] = i;
      }
      model.nodesWereInserted(this,indices);
   }

   public void onAdd(Object substance, CollectionItem collection, 
                     Object value, Object added, Object param) {
      Log.trace("gui.collectionUpdate",
                "onAdd "+collection.getLongName()+" "+added+" - "+areChildrenUptodate);
      ObjectNode newNode;
      if (areChildrenUptodate) {
         newNode = new ObjectNode(model,added,substance,collection); 
         newNode.setLeaf(TreeModel.isLeafNode(model,newNode,added,showRelations));
         add(newNode);
         model.nodesWereInserted(this,new int[] {getChildCount()-1});
         setChildrenUptodate(true);
      } else {
         TreeModel.addNodes(model,this,getUserObject(),showRelations);
         int[] indices = new int[getChildCount()];
         for(int i=0; i<indices.length; i++) {
            indices[i] = i;
         }
         newNode = findNode(collection,added);
         model.nodesWereInserted(this,indices);
      }
      if (newNode!=null) {
         model.setSelection(new TreePath(newNode.getPath()));
      }
   }

   public void onRemove(Object substance, CollectionItem collection, 
                        Object value, Object removed, Object param) {
      Log.trace("gui.collectionUpdate","onRemove "+collection.getLongName()+" "+removed);
      onChange(substance,collection,value,param);
   }

   public void updateChildren() {
      if (!areChildrenUptodate()) {
         Log.trace("gui.events","updateChildren "+this);
         TreeModel.addNodes(model,this,getUserObject(),showRelations);
         int[] indices = new int[getChildCount()];
         for(int i=0; i<indices.length; i++) {
            indices[i] = i;
         }
         model.nodesWereInserted(this,indices);
         setChildrenUptodate(true);
      } else {
         Log.trace("gui.events","children are uptodate for "+this);
      }
   }

   public String toString() {
      return substance+"."+relation+" -> "+getUserObject();
   }
}

