/*
  Copyright (C) 2002-2003 Laurent Martelli <laurent@aopsys.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.gui.web;

import org.objectweb.jac.aspects.gui.*;
import org.objectweb.jac.aspects.gui.FieldView;
import org.objectweb.jac.core.Collaboration;
import org.objectweb.jac.core.Naming;
import org.objectweb.jac.core.Wrappee;
import org.objectweb.jac.core.rtti.ClassItem;
import org.objectweb.jac.core.rtti.ClassRepository;
import org.objectweb.jac.core.rtti.CollectionItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.core.rtti.MemberItem;
import org.objectweb.jac.core.rtti.MetaItem;
import org.objectweb.jac.core.rtti.MethodItem;
import org.objectweb.jac.util.Classes;
import org.objectweb.jac.util.ExtArrays;
import org.objectweb.jac.util.Log;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

public class Table extends AbstractCollection
   implements HTMLViewer, TableListener
{
   HTMLViewer[] cellViewers;
   ExtendedTableModel tableModel;

   public Table(ViewFactory factory, DisplayContext context,
                CollectionItem collection, Object substance,
                ExtendedTableModel model, 
                org.objectweb.jac.aspects.gui.CollectionItemView itemView) {
      super(factory,context,collection,substance,model,itemView);
      this.tableModel = model;      
      if (model!=null) {
         this.cellViewers = new HTMLViewer[tableModel.getHeaders().length];
         setCellRenderers();
      }
   }

   public void setColumnsInfo(String[] headers, FieldItem[] fields, 
                              ClassItem[] classes, ClassItem[] viewerClasses) {
      if (!(headers.length==classes.length && 
            classes.length==viewerClasses.length &&
            viewerClasses.length==fields.length)) {
         throw new RuntimeException("headers, fields, classes and "+
                                    "viewerClasses must be the same size");
      }
      this.cellViewers = new HTMLViewer[headers.length];
   }

   protected void setCellRenderers() {
      MemberItem[] members = tableModel.getMembers();
      String[] headers = tableModel.getHeaders();

      for (int i=0; i<members.length; i++) {
         if (members[i] instanceof FieldItem) {
            cellViewers[i] = 
               (HTMLViewer)tableModel.getCellRenderer(this,i,factory,context);
         } else if (members[i] instanceof MethodItem) {
            MethodItem method = (MethodItem)members[i];
            try {
               cellViewers[i] = (HTMLViewer)factory.createView(
                  method.getName(),
                  "Method",
                  new Object[] {substance,method},context);
            } catch (Exception e) {
               Log.error("Failed to instanciate TableCellRenderer "+
                         method.getName()+" for column "+headers[i]);
               e.printStackTrace();
            }
         }
         if (cellViewers[i] instanceof FieldView)
            ((FieldView)cellViewers[i]).setAutoUpdate(false);
         if (cellViewers[i] instanceof CollectionView)
            ((CollectionView)cellViewers[i]).setEditor(false);
      }      
   }

   public void sort() {
      ((TableSorter)model).sort(this);
   }

   /**
    * Sorts the collection.
    *
    * @param column the index of the column used to sort
    */
   public void sort(int column) {
      ((TableSorter)model).sortByColumn(column, true);
   }

   // HTMLViewer interface

   public void genHTML(PrintWriter out) throws IOException {
      MethodItem adder = collection.getAdder();
      boolean embeddedAdder = 
         adder!=null && GuiAC.getView(adder,itemView.getName()).isEmbedded() && 
         GuiAC.isAutoCreate(collection);
      if (model.getRowCount()==0 && !embeddedAdder) {
         out.println("none");
      }

      genAdderAndRemover(out);

      if (model.getRowCount()==0 && !embeddedAdder) {
         return;
      }

      Log.trace("gui.table",itemView+".Embedded editors="+itemView.isEmbeddedEditors());

      boolean nonViewable = GuiAC.isViewableItems(collection);

      String[] headers = tableModel.getHeaders();

      out.println("<table class=\"table\">");
      out.println("  <thead>");
      out.println("    <tr>");
      if (showRowNumbers) {
         out.println("      <th style=\"width:"+
                     ((""+model.getRowCount()).length())+"ex\" class=\"empty\"></th>");
      }
      for (int i=0; i<headers.length; i++) {
         out.print("      <th>");
         if (headers[i]!=null) {
            out.print(sortLink(i,headers[i]));
         }
         out.println("      </th>");
      }
      if (!nonViewable)
         out.println("      <th style=\"width:0*\" class=\"empty\"></th>"); // view button
      if (GuiAC.isRemovable(collection))
         out.println("      <th style=\"width:0*\" class=\"empty\"></th>"); // remove button
      out.println("    </tr>");
      out.println("  </thead>");

      boolean bodyOpened = false;

      Object defaultsObject = null;
   
      // Default value editors in the table
      if (GuiAC.hasEditableDefaultValues(collection) && embeddedAdder) {
         try {
            ClassItem componentType = collection.getComponentType();
            defaultsObject = Naming.getObject("defaults."+collection.getLongName());
            if (defaultsObject==null) {
               Naming.setName("defaults."+collection.getLongName());
               defaultsObject = componentType.newInstance();
               clearDefaultEditors();
            }
            if (!bodyOpened) {
               out.println("  <tbody class=\"defaultObject\">");
               bodyOpened = true;
            }
            out.println("    <tr class=\"vspace\"></tr>");
            out.println("    <tr class=\"defaultObject\">");
            FieldItem[] defaultsFields = GuiAC.getDefaultsAttributesOrder(componentType);
            if (showRowNumbers) {
               out.println("      <td></td>");
            }
            for (int col=0; col<tableModel.getColumnCount(); col++) {
               out.print("      <td>");
               MemberItem member = tableModel.getMembers()[col];
               if (defaultsFields!=null && !ExtArrays.contains(defaultsFields,member)) {
                  out.println("</td>");
                  continue;
               }
               GuiAC.pushGraphicContext(member);
               try {
                  if (member instanceof FieldItem && 
                      !(member instanceof CollectionItem)) 
                  {
                     MethodItem setter = ((FieldItem)member).getSetter();
                     if (GuiAC.isCreationAttribute((FieldItem)member) &&
                         setter!=null) {
                        // <BAD>WE SHOULD ONLY BUILD THE EDITORS ONCE</BAD>
                        FieldEditor editor = GenericFactory.getEditorComponent(
                           factory, context, defaultsObject, setter, 0, true, null);
                        defaultsEditors.add(editor);
                        context.addEditor(editor);
                        ((HTMLViewer)editor).genHTML(out);
                        editor.setEmbedded(true);
                        //editorContainer.addEditor(editor);
                        //container.addView(editor);
                     }
                  } else if (member instanceof CollectionItem) {
                     if (GuiAC.isCreationAttribute((FieldItem)member)) {
                        CollectionItem collection = (CollectionItem)member;
                        HTMLViewer view = (HTMLViewer)
                           GenericFactory.getCollectionPane(
                              factory,context,defaultsObject,null,null,collection);
                        view.genHTML(out);
                     }
                  }
               } finally {
                  GuiAC.popGraphicContext();
               }
               out.println("</td>");
            }
            out.println("      <td>"+eventURL("set","onSetDefaults","")+"</td>");
            out.println("    </tr>");
            out.println("    <tr class=\"vspace\"></tr>");
            if (bodyOpened) {
               out.println("  </tbody>");
               bodyOpened = false;
            }
         } catch (Exception e) {
            e.printStackTrace();
         }
      }
   
      MemberItem[] members = tableModel.getMembers();
      for (int index=startIndex; 
           (!split || index<startIndex+rowsPerPage) && index<model.getRowCount(); 
           index++) 
      {
         Object substance = tableModel.getObject(index);
         if (!bodyOpened) {
            out.println("  <tbody>");
            bodyOpened = true;
         }
         String events="ondblclick=\"viewTableRow()\"";
         if (selected==index)
            out.println("    <tr class=\"selected\">");
         else
            out.println("    <tr class=\""+
                        ((index%2==0)?"even":"odd")+"\">");
         if (showRowNumbers) {
            out.println("      <td>"+(index+1)+"</td>");
         }
         for (int col=0; col<tableModel.getColumnCount(); col++) {
            out.print("      <td>");
            try {
               HTMLViewer viewer = null;
               if (itemView.isEmbeddedEditors()) {
                  if (members[col] instanceof FieldItem)
                     viewer = 
                        getFieldEditor((FieldItem)members[col],substance);
               }
               if (viewer!=null) {
                  viewer.genHTML(out);
               } else {
                  if (cellViewers[col] instanceof Method) {
                     out.println("      <a href=\""+eventURL("onTableInvoke")+
                                 "&amp;index="+index+"&amp;method="+
                                 (members[col]).getName()+"\">"+
                                 (GuiAC.getLabel(members[col]))+"</a>");
                  
                  } else {
                     FieldView cellViewer = (FieldView)cellViewers[col];
                     if (cellViewer!=null) {
                        if (cellViewer instanceof TableCellViewer) {
                           ((TableCellViewer)cellViewer).setRow(index);
                           ((TableCellViewer)cellViewer).setColumn(col);
                        }
                        if (cellViewer instanceof CollectionView) 
                           ((CollectionView)cellViewer).updateModel(substance);
                        else {
                           cellViewer.setSubstance(substance);
                           cellViewer.setValue(tableModel.getValueAt(index,col));
                        }
                        if (cellViewer instanceof ReferenceView) {
                           ((ReferenceView)cellViewer).setEventURL(
                              eventURL("onCellSelection")+"&amp;row="+index+"&amp;col="+col);
                        }
                        ((HTMLViewer)cellViewer).genHTML(out);
                     } else {
                        Object value = tableModel.getValueAt(index,col);
                        if (value!=null) {
                           out.print(GuiAC.toString(value));
                        }
                     }
                  }
               }
            } catch (Exception e) {
               Log.error("Failed to genHTML for cell["+
                         tableModel.getHeaders()[col]+","+index+"]: "+e);
               e.printStackTrace();
            }
            out.println("</td>");
         }
         if (!nonViewable)
         {
            out.print("      <td>");
            if (collection.getAttribute(GuiAC.NEW_WINDOW)!=null) {
               out.print("<a target=\""+collection.getName()+"\""+
                         " href=\""+eventURL("onView")+
                         "&amp;index="+index+"&amp;\">"+
                         "details</a>");
            } else {
               out.print("<a href=\""+eventURL("onView")+
                         "&amp;index="+index+"\">details</a>");
            }
            out.println("</td>");
         }
         if (GuiAC.isRemovable(collection)) {
            out.print("<td>");
            if (isEditor)
               out.print(removeLink(index));
            out.println("</td>");
         }
         out.println("    </tr>");
      }
      if (bodyOpened) {
         out.println("  </tbody>");
         bodyOpened = false;
      }

      // Embedded adder in the table
      String focusId = null; // Id of HTML element to focus (if any)
      if (GuiAC.isAddable(substance,collection)) {
         try {
            ClassItem componentType = collection.getComponentType();
            if (embeddedAdder) {
               if (addedObject==null) {
                  addedObject = componentType.newInstance();
                  EventHandler.initAutocreatedObject(
                     addedObject,substance,collection);
                  initAddedObject(addedObject,defaultsObject);
                  clearEmbeddedEditors();
               }
               out.println("  <tbody>");
               out.println("    <tr class=\"vspace\"></tr>");
               out.println("    <tr class=\"addedObject\""+
                           " onkeypress=\"return commitFormOnEnter(event.which,this,"+
                           "'event=onAddEmbedded&amp;source="+getId()+"')\">");
               if (showRowNumbers) {
                  out.println("      <td></td>");
               }
               for (int col=0; col<tableModel.getColumnCount(); col++) {
                  out.print("      <td>");
                  MemberItem member = tableModel.getMembers()[col];
                  GuiAC.pushGraphicContext(member);
                  try {
                     if (member instanceof FieldItem && 
                         !(member instanceof CollectionItem))
                     {
                        MethodItem setter = ((FieldItem)member).getSetter();
                        if (GuiAC.isCreationAttribute((FieldItem)member) && 
                            setter!=null) {
                           FieldEditor editor = GenericFactory.getEditorComponent(
                              factory, context, addedObject, setter, 0, true, null);
                           if (focusId==null)
                              focusId = editor.getLabel();
                           embeddedEditors.add(editor);
                           ((HTMLViewer)editor).genHTML(out);
                           editor.setEmbedded(true);
                           context.addEditor(editor);
                           //editorContainer.addEditor(editor);
                           //container.addView(editor);
                        }
                     } else if (member instanceof CollectionItem) {
                        if (GuiAC.isCreationAttribute((FieldItem)member)) {
                           CollectionItem coll = (CollectionItem)member;
                           CollectionItem index = (CollectionItem)
                              coll.getComponentType().getAttribute(GuiAC.INDEXED_FIELD_SELECTOR);
                           if (index==null) {
                              HTMLViewer view = (HTMLViewer)
                                 GenericFactory.getCollectionPane(
                                    factory,context,addedObject,null,null,coll);
                              view.genHTML(out);
                           } else {
                              FieldEditor editor = factory.createEditor(
                                 "editor "+Naming.getName(substance)+"."+coll.getName(),
                                 "IndicesSelector",
                                 new Object[] {coll,addedObject,
                                               new ListModel(coll,addedObject),
                                               GuiAC.getView(coll,itemView.getName())},
                                 context);
                              embeddedEditors.add(editor);
                              ((HTMLViewer)editor).genHTML(out);
                              editor.setEmbedded(true);
                              context.addEditor(editor);
                           }
                        }
                     }
                  } catch(Exception e) {
                     Log.error("Failed to gen HTML for embedded object to add, column "+
                               tableModel.getHeaders()[col]);
                     e.printStackTrace();
                  } finally {
                     GuiAC.popGraphicContext();
                  }
                  out.println("</td>");
               }
               out.println("      <td>"+eventURL(GuiAC.getLabelAdd(),"onAddEmbedded","")+"</td>");
               out.println("    </tr>");
            }
         } catch (Exception e) {
            e.printStackTrace();
         }
      }

      out.println("  </tbody>");
      
      out.println("</table>");

      if (focusId!=null) {
         out.println("<script type=\"text/javascript\">"+
                     "element=getElementById('"+focusId+"');"+
                     "element.focus();"+
                     "window.scroll(0,10000);"+
                     "</script>");
      }
   }

   /**
    * Returns a cell editor for a field of an object. Editors are
    * cached, so you'll always get the same object for the same field
    * and substance, unless clearCellEditors() is called.
    * @see #clearCellEditors()
    */
   protected HTMLViewer getFieldEditor(FieldItem field, Object substance) {
      HTMLViewer editor = null;
      Map fieldEditors = (Map)cellEditors.get(field);
      if (fieldEditors!=null) {
         editor = (HTMLViewer)fieldEditors.get(substance);
      } else {
         fieldEditors = new HashMap();
         cellEditors.put(field,fieldEditors);
      }
      if (editor==null) {
         if (!(field instanceof CollectionItem))
         {
            MethodItem setter = field.getSetter();
            if (setter!=null) {
               editor = (HTMLViewer)GenericFactory.getEditorComponent(
                  factory, context, field.getSubstance(substance), setter, 0, true, null);
            }
         } else  {
            CollectionItem coll = (CollectionItem)field;
            CollectionItem index = (CollectionItem)
               coll.getComponentType().getAttribute(GuiAC.INDEXED_FIELD_SELECTOR);
            if (index==null) {
               editor = (HTMLViewer)GenericFactory.getCollectionPane(
                  factory,context,substance,null,itemView,coll);
            } else {
               editor = (HTMLViewer)factory.createEditor(
                  "editor "+Naming.getName(substance)+"."+coll.getName(),
                  "IndicesSelector",
                  new Object[] {coll,substance,
                                new ListModel(coll,substance),
                                GuiAC.getView(coll,itemView.getName())},
                  context);
            }
         }
         fieldEditors.put(substance,editor);
         if (editor instanceof FieldEditor) {
            ((FieldEditor)editor).setEmbedded(true);
            context.addEditor((FieldEditor)editor);
         }
         if (editor instanceof ReferenceEditor) {
            ((ReferenceEditor)editor).setEditable(false);
         }
      }
      Log.trace("table","editor["+field.getName()+","+substance+"] -> "+editor);
      return editor;
   }

   /**
    * Removes editors of embedded added object
    */
   protected void clearEmbeddedEditors() {
      Iterator it = embeddedEditors.iterator();
      while (it.hasNext()) {
         FieldEditor editor = (FieldEditor)it.next();
         editor.close();
         context.removeEditor(editor);
      }
      embeddedEditors.clear();
   }

   /**
    * Removes editors of embedded added object
    */
   protected void clearDefaultEditors() {
      Iterator it = defaultsEditors.iterator();
      while (it.hasNext()) {
         FieldEditor editor = (FieldEditor)it.next();
         context.removeEditor(editor);
      }
      defaultsEditors.clear();
   }

   // FieldItem -> (Object -> FieldEditor)
   Hashtable cellEditors = new Hashtable();

   /**
    * Removes editors of embedded added object
    */
   protected void clearCellEditors() {
      Iterator it = cellEditors.values().iterator();
      while (it.hasNext()) {
         Iterator editors = ((Map)it.next()).values().iterator();
         while (editors.hasNext()) {
            View editor = (View)editors.next();
            if (editor!=null)
               editor.close();
            context.removeEditor(editor);
         }
      }
      cellEditors.clear();
   }

   public void close() {
      super.close();
      clearEmbeddedEditors();
      clearDefaultEditors();
      clearCellEditors();
   }

   /**
    * Initialize fields of added object from one of defaultsObject
    */
   protected void initAddedObject(Object addedObject, Object defaultsObject) {
      if (defaultsObject==null || addedObject==null) {
         return;
      }
      for (int col=0; col<tableModel.getColumnCount(); col++) {
         MemberItem member = tableModel.getMembers()[col];
         if (member instanceof FieldItem && !(member instanceof CollectionItem)) {
            FieldItem field = (FieldItem)member;
            MethodItem setter = field.getSetter();
            if (GuiAC.isCreationAttribute(field) && setter!=null) {
               Object value = field.getThroughAccessor(defaultsObject);
               try {
                  field.set(addedObject,value);
               } catch (Exception e) {
                  Log.error("Failed to set default value for field "+field+": "+e);
                  e.printStackTrace();
               }
            }
         }
      }
   }

   public void onCellSelection(int row, int col) {
      MemberItem[] members = tableModel.getMembers();
      EventHandler.get().onSelection(context,(FieldItem)members[col],
                                     tableModel.getObject(row,col),
                                     null,null,true);
   }

   List embeddedEditors = new Vector();
   Object addedObject = null;

   public void onEmbeddedAddToCollection() {
      JacRequest request = WebDisplay.getRequest();
      Log.trace("gui.events",
                "onEmbeddedAddToCollection "+collection.getName()+" "+request);
      for (int i=0; i<embeddedEditors.size(); i++) {
         FieldEditor editor = (FieldEditor)embeddedEditors.get(i);
         editor.close();
      }
      EventHandler.get().onInvoke(
         context,substance,collection.getAdder(),false,
         new Object[] {addedObject},null,null);
      clearEmbeddedEditors();
      /**
         // onView(addedObject); 

         It does not work because 

            a) onInvoke() above is asynchronous and starts its own
               thread, so we are not sure addedObject is added
               "completely" (the table model may not be up to date for
               instance)

            b) InvokeThread already calls WebDisplay.refresh(), so
               calling it again in onView() will fail

      */
      addedObject = null;
   }
   
   List defaultsEditors = new Vector();
   
   public void onSetDefaults() {
      Log.trace("gui.events","onSetDefaults "+collection.getName());
      initAddedObject(addedObject,
                      Naming.getObject("defaults."+collection.getLongName()));
      context.getDisplay().refresh();
   }
   
   public void onHeaderClick(int column) {
      //      sort(column);
      TableSorter sorter = ((TableSorter)model);
      SortCriteria criteria = sorter.getSortCriteria(column);
      if (criteria!=null) {
         if (criteria.isAscending()) {
            criteria.toggleAscending();
            sorter.sortByColumn(column,criteria.isAscending());
         } else {
            sorter.sortByColumn(-1,criteria.isAscending());            
         }
      } else {
         sorter.sortByColumn(column,true);
      }
//       sortColumn = column;
//       ((TableSorter)model).sortByColumn(column, true); 
      context.getDisplay().refresh();
   }

   public View onRowEvent(int row, int col) {
      FieldView cellViewer = 
         (FieldView)tableModel.getCellRenderer(this,col,factory,context);
      cellViewer.setValue(tableModel.getValueAt(row,col));
      return cellViewer;
   }

}
