/*
  Copyright (C) 2002-2003 Renaud Pawlak <renaud@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.integrity;

import org.objectweb.jac.core.AspectComponent;
import org.objectweb.jac.core.rtti.ClassItem;
import org.objectweb.jac.core.rtti.CollectionItem;
import org.objectweb.jac.core.rtti.FieldItem;
import org.objectweb.jac.core.rtti.MethodItem;
import org.objectweb.jac.core.rtti.RttiAC;
import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Strings;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

/**
 * This aspect handle different kinds of data integrity among objects
 * sets. */

public class IntegrityAC extends AspectComponent implements IntegrityConf {

   HashMap preConditionsFields = new HashMap();
   HashMap postConditionsFields = new HashMap();

   RoleWrapper roleWrapper = new RoleWrapper(this);

   public static final String CONSTRAINTS = "IntegrityAC.CONSTRAINTS";

   public static final String DISABLE_ADD_TO_REPOSITORY = "IntegrityAC.DISABLE_ADD_TO_REPOSITORY";

   HashSet classesWithAssociations = new HashSet();

   public void declareAssociation(ClassItem classA, String fieldNameA, 
                                  ClassItem classB, String fieldNameB)
   {
      Log.trace("integrity","declareAssociation "+
                classA.getName()+"."+fieldNameA+" <-> "+
                classB.getName()+"."+fieldNameB);
      classesWithAssociations.add(classA);
      classesWithAssociations.add(classB);
      FieldItem fieldA = classA.getField(fieldNameA);
      FieldItem fieldB = classB.getField(fieldNameB);
      fieldA.setAttribute(RttiAC.OPPOSITE_ROLE,fieldB);
      fieldB.setAttribute(RttiAC.OPPOSITE_ROLE,fieldA);
   }

   public void declareDirectionalAssociation(
      ClassItem classA, String fieldNameA, 
      ClassItem classB, String fieldNameB)
   {
      Log.trace("integrity","declareDirectionalAssociation "+
                classA.getName()+"."+fieldNameA+" -> "+
                classB.getName()+"."+fieldNameB);
      classesWithAssociations.add(classA);
      FieldItem fieldA = classA.getField(fieldNameA);
      FieldItem fieldB = classB.getField(fieldNameB);
      fieldA.setAttribute(RttiAC.OPPOSITE_ROLE,fieldB);
   }

   public void declareRepository(String repositoryName,
                                 CollectionItem collection, 
                                 FieldItem field) {      
      RepositoryWrapper rw = new RepositoryWrapper(this,repositoryName,
                                                   collection,field);
      Log.trace("integrity.repository","declareRepository "+repositoryName+
                " "+collection+" "+field);
      if (field instanceof CollectionItem) {
         pointcut("ALL",field.getParent().getName(),
                  "ADDERS("+field.getName()+")",
                  rw,"addToRepository",null);
      } else {
         pointcut("ALL",field.getParent().getName(),
                  "SETTERS("+field.getName()+")",
                  rw,"addToRepository",null);         
      }
      
      // This pointcut adds an attribute in context so that we can
      // avoid adding an object twice in the repository
      pointcut(repositoryName,collection.getParent().getName(),
               "ADDERS("+collection.getName()+")",
               rw,"doAddToRepository",null);
   }

   public void declareConstraint(ClassItem cl, String relationName,
                                 FieldItem target, String constraint) {
      FieldItem relation = cl.getField(relationName);
      Log.trace("integrity.constraint","declareConstraint "+
                relation.getLongName()+"->"+target.getLongName()+": "+constraint);
      List constraints = (List)target.getAttributeAlways(CONSTRAINTS);     
      if (constraints==null) {
         constraints = new Vector();
         target.setAttribute(CONSTRAINTS,constraints);
      }
      int c;
      if (constraint.compareToIgnoreCase("DELETE_CASCADE")==0) {
         Log.warning("IntegrityAC","DELETE_CASCADE is not implemented yet");
         c = Constraint.DELETE_CASCADE;
      } else if (constraint.compareToIgnoreCase("SET_NULL")==0) {
         c = Constraint.SET_NULL; 
      } else if (constraint.compareToIgnoreCase("FORBIDDEN")==0) {
         c = Constraint.FORBIDDEN;
      } else {
         throw new RuntimeException("Unknown constraint type "+constraint);
      }
      constraints.add(new Constraint(relation,c));
      if (target instanceof CollectionItem) {
         pointcut("ALL",target.getParent().getName(),
                  "REMOVERS("+target.getName()+")",
                  new RepositoryWrapper(this,"",(CollectionItem)target,null),
                  "removeFromRepository",null);
      } else {
         pointcut("ALL",target.getParent().getName(),
                  "SETTERS("+target.getName()+")",
                  new RepositoryWrapper(this,"",null,target),
                  "removeFromRepository",null);
      }
   }

   boolean hasConditions=false;

   private void addCondition(ClassItem cl,
                             String fieldName,
                             MethodItem constraint,
                             Object[] params,
                             String errorMsg,
                             HashMap conditionsList) {
      hasConditions=true;
      FieldItem field = cl.getField(fieldName);
      field.setAttribute(RttiAC.CONSTRAINTS, Boolean.TRUE);

      Vector vect = (Vector) conditionsList.get(field);
      if (vect == null)
         vect = new Vector();

      ConstraintDescriptor obj = new ConstraintDescriptor(
         constraint,
         params,
         errorMsg);
      vect.add(obj);
      conditionsList.put(field, vect);
   }
   
   public void addPostCondition(ClassItem cl,
                                String fieldName,
                                MethodItem constraint,
                                Object[] params,
                                String errorMsg) {
      addCondition(cl,
                   fieldName,
                   constraint,
                   params,
                   errorMsg,
                   postConditionsFields);
   }

   public void addPreCondition(ClassItem cl,
                               String fieldName,
                               MethodItem constraint,
                               Object[] params,
                               String errorMsg) {
      addCondition(cl,
                   fieldName,
                   constraint,
                   params,
                   errorMsg,
                   preConditionsFields);
   }

   /**
    * Activates constraints checking.
    *
    * <p>This method must be called once after all calls to
    * <code>addPreCondition</code> and <code>addPostCondition</code>
    * to effectively add pointcuts for these methods.</p>
    *
    * @see #addPreCondition(ClassItem,String,MethodItem,Object[],String)
    * @see #addPostCondition(ClassItem,String,MethodItem,Object[],String)
    */

   protected void doCheck() {
      ConstraintWrapper cw = new ConstraintWrapper(this);
      pointcut("ALL",
               "ALL",
               "SETTERS(<"+RttiAC.CONSTRAINTS+">)",
               cw,
               "testConditions",
               null);
   }

   /**
    * Activates primary keys checking.
    *
    * <p>This method must be called once in the file integrity.acc to
    * effectively add pointcuts for primary key checking methods.</p>
    *
    * @see org.objectweb.jac.core.rtti.RttiAC#definePrimaryKey(ClassItem,String,String[])
    */

   void doPrimaryKeyCheck() {
      PrimaryKeyWrapper pw = new PrimaryKeyWrapper(this);
      pointcut("ALL",
               "ALL",
               "ADDERS(<"+RttiAC.PRIMARY_KEY+">)",
               pw,
               "checkDoubles",
               null);
   }

   public void whenConfigured() {
      Log.trace("integrity",this+".whenConfigured");
      if (hasConditions) {
         doCheck();
      }
      doPrimaryKeyCheck();
      String classExpr = Strings.join(classesWithAssociations," || ");
      pointcut("ALL",classExpr,
               "ADDERS(<"+RttiAC.OPPOSITE_ROLE+">) "+
               "|| REMOVERS(<"+RttiAC.OPPOSITE_ROLE+">) "+
               "|| SETTERS(<"+RttiAC.OPPOSITE_ROLE+">)",
               roleWrapper,"updateOppositeRole",null);
      pointcut("ALL",classExpr,"CONSTRUCTORS",
               roleWrapper,"initAutoCreatedObject",null);
   }

}
