/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.aspects.persistence;


import org.objectweb.jac.aspects.naming.NameGenerator;
import org.objectweb.jac.core.rtti.*;
import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Semaphore;
import org.objectweb.jac.util.Strings;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StreamTokenizer;
import java.lang.reflect.*;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;

/**
 * A FileSystem storage
 */

public class FSStorage implements Storage {

   // names -> oid 
   private Hashtable names = new Hashtable();
   // oid -> name
   private Hashtable oids = new Hashtable();
   // oid -> classid
   private Hashtable classes = new Hashtable();
   
   private File basedir;
   private File oidsFile;
   private File classesFile;

   private long lastOID = 0;

   Semaphore semaphore = new Semaphore(1);

   NameGenerator nameGen;
   File nameCountersFile;

   /**
    * Create a new file system storage
    * @param basedirName name of the directory where to store data files
    */
   public FSStorage(String basedirName) {
      Log.trace("storage","new FSStorage "+basedirName);
      basedir = new File(basedirName);
      if (!basedir.isAbsolute()) {
         basedir = new File(org.objectweb.jac.core.Jac.getJacRoot(),basedir.toString());
      }
      if (! basedir.isDirectory()) {
         basedir.mkdirs();
      }
      oidsFile = new File(basedir,"oids");
      try {
         oidsFile.createNewFile();
         readOids();
      } catch(Exception e) {
         e.printStackTrace();
      }

      classesFile = new File(basedir,"classes");
      try {
         classesFile.createNewFile();
         readClasses();
      } catch(Exception e) {
         e.printStackTrace();
      }

      nameGen = new NameGenerator();

      nameCountersFile = new File(basedir,"nameCounters");
      try {
         nameCountersFile.createNewFile();
         readNameCounters();
      } catch(Exception e) {
         e.printStackTrace();
      }
      
   }

   /**
    * Safely close the storage. Waits for all operations to terminate.
    */
   public void close() {
      // wait for completion of deleteObject
      semaphore.acquire();
      System.out.println("FSStorage shutdown hook completed");
   }

   /**
    * Read the "oids" file
    */
   void readOids() throws Exception 
   {
      StreamTokenizer tokens = getStreamTokenizer(oidsFile);

      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         Log.trace("storage",2,
                   "token : "+tokens.sval+"/"+tokens.nval+" "+tokens.ttype);
         LongOID oid = new LongOID(Long.parseLong(tokens.sval));
         long oidval = oid.getOID();
         if (oidval>lastOID)
            lastOID = oidval;
         tokens.nextToken();
         Log.trace("storage",2,
                   "token : "+tokens.sval+"/"+tokens.nval+" "+tokens.ttype);
         String name = tokens.sval;
         oids.put(oid,name);
         names.put(name,oid);
      }
   }

   /**
    * Writer the "oids" file
    */ 
   void writeOids() throws Exception 
   {
      Log.trace("storage","writeOids");
      writeMapToFile(oids,oidsFile);
   }

   /**
    * Read the "nameCounters" file
    */
   void readNameCounters() throws IOException, FileNotFoundException
   {
      StreamTokenizer tokens = getStreamTokenizer(nameCountersFile);
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         String className = tokens.sval;
         tokens.nextToken();
         Long counter = new Long(tokens.sval);
         Log.trace("storage",2,className+" -> "+counter);
         nameGen.put(className,counter);
      }      
   }

   /**
    * Write the "nameCounters" file
    */
   void writeNameCounters() throws IOException {
      Log.trace("storage","writeNameCounters");
      writeMapToFile(nameGen,nameCountersFile);
   }

   /**
    * Write the content of a Map to a File, using the toString() of
    * the keys and values.
    * @param map the map
    * @param file the file */
   void writeMapToFile(Map map, File file) throws IOException {
      PrintWriter writer = 
         new PrintWriter(new FileWriter(file.toString(),false),false);

      Iterator it = map.entrySet().iterator();
      while (it.hasNext()) {
         Map.Entry entry = (Map.Entry)it.next();
         writer.println(entry.getKey().toString()+" "+entry.getValue());         
      }
      writer.flush();
   }

   void readClasses() throws Exception 
   {
      Log.trace("storage","readClasses");
      StreamTokenizer tokens = getStreamTokenizer(classesFile);
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         LongOID oid = new LongOID(Long.parseLong(tokens.sval));
         long oidval = oid.getOID();
         if (oidval>lastOID)
            lastOID = oidval;
         tokens.nextToken();
         String classid = tokens.sval;
         Log.trace("storage",2,oid.toString()+" -> "+classid);
         classes.put(oid,classid);
      }
   }

   void writeClasses() throws Exception 
   {
      Log.trace("storage","writeClasses");
      writeMapToFile(classes,classesFile);
   }

   protected StreamTokenizer getStreamTokenizer(File file) 
      throws FileNotFoundException, IOException
   {
      file.createNewFile();
      StreamTokenizer tokens = new StreamTokenizer(new FileReader(file));
      tokens.resetSyntax();
      tokens.wordChars('\000','\377');
      tokens.whitespaceChars(' ',' ');
      tokens.whitespaceChars('\t','\t');
      tokens.whitespaceChars('\n','\n');
      tokens.whitespaceChars('\r','\r');
      return tokens;
   }

   public OID createObject(String className) throws Exception
   {
      semaphore.acquire();
      try {
         lastOID++;
         LongOID oid = new LongOID(lastOID);
         classes.put(oid,className);
         PrintWriter writer = 
            new PrintWriter(new FileWriter(classesFile.toString(),true),true);
         writer.println(oid.toString()+" "+className);
         return oid;
      } finally {
         semaphore.release();
      }
   }

   public void deleteObject(OID oid) throws Exception
   {
      semaphore.acquire();
      try {
         Log.trace("storage","deleteObject("+oid+")");
         String name = (String)oids.get(oid);
         if (name!=null) {
            names.remove(name);
            oids.remove(oid);
            writeOids();
         } else {
            Log.warning("storage","FSStorage.deleteObject: oid "+
                        oid+" does not have name");
            Log.trace("storage","names = "+classes);
         }
         if (classes.containsKey(oid)) {
            classes.remove(oid);
            writeClasses();
         } else {
            Log.warning("storage","FSStorage.deleteObject: oid "+
                        oid+" does not have a class");
            Log.trace("storage","classes = "+classes);
         }
      } catch(Exception e) {
         e.printStackTrace();
      } finally {
         Log.trace("storage","deleteObject("+oid+") DONE");
         semaphore.release();
      }
      // *** TODO: WE SHOULD ALSO REMOVE COLLECTIONS OWNED BY THE
      // ***       DELETED OBJECT
   }

   public void setField(OID oid, FieldItem field, Object value) 
      throws Exception
   {
      Log.trace("storage","setField("+oid+","+field+","+value+")");
      Properties props = new Properties();
      File objectFile = new File(basedir,oid.toString());
      objectFile.createNewFile();
      props.load(new FileInputStream(objectFile));
      props.setProperty(field.getName(),ValueConverter.objectToString(value));
      props.store(new FileOutputStream(objectFile),getClassID(oid)+" "+oid);
   }

   public void updateField(OID oid, FieldItem field, Object value) 
      throws Exception
   {
      Log.trace("storage","updateField("+oid+","+field+","+value+")");
      setField(oid,field,value);
   }

   /** oid -> Properties */
   Hashtable cache = new Hashtable();

   public Object getField(OID oid, FieldItem field)
      throws Exception
   {
      Log.trace("storage","getField("+oid+","+field+")");
      Object ret = null;
      Properties props = (Properties)cache.get(oid);
      if (props==null) {
         props = new Properties();
         File objectFile = new File(basedir,oid.toString());
         props.load(new FileInputStream(objectFile));
         cache.put(oid,props);
      }
      String value = (String)props.get(field.getName());
      if (value != null) {
         ret = ValueConverter.stringToObject(value);
      } else {
         if (field.isPrimitive()) {
            Log.warning("no such field in storage "+oid+","+field.getName());
         }
         ret = null;
      }
      Log.trace("storage",2," -> "+ret);
      return ret;
   }

   protected Properties getFields(OID oid) throws IOException
   {
      Properties props = new Properties();
      props.load(new FileInputStream(new File(basedir,oid.toString())));
      return props;
   }

   public StorageField[] getFields(OID oid,ClassItem cl,FieldItem[] fields) 
      throws Exception
   {
      Properties props = getFields(oid);
      StorageField ret[] = new StorageField[fields.length];
      int count = 0;
      for (int i=0; i<fields.length; i++) {
         String value = (String)props.get(fields[i].getName());
         if (value!=null)
            ret[i] = new StorageField(cl,fields[i],ValueConverter.stringToObject(value));
         else
            ret[i] = new StorageField(cl,fields[i],null);
      }
      return ret;
   }

   // Collection methods

   protected long getCollectionSize(OID cid) throws Exception
   {
      return getList(cid).size();
   }

   public OID getCollectionID(OID oid, CollectionItem collection) 
      throws Exception
   {
      return (OID)getField(oid,collection);
   }

   // List methods

   public void clearList(OID cid) throws Exception
   {
      saveList(cid,new Vector());
   }

   public List getList(OID oid, CollectionItem collection)
      throws Exception
   {
      Log.trace("storage","getList("+oid+","+collection+")");
      return getList(getCollectionID(oid,collection));
   }

   public List getList(OID cid)
      throws Exception
   {
      Log.trace("storage","getList("+cid+")");
      File file = new File(basedir,cid.toString());
      StreamTokenizer tokens = getStreamTokenizer(file);
      Vector ret = new Vector();
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         try {
            ret.add(ValueConverter.stringToObject(Strings.unslashify(tokens.sval)));
         } catch (Throwable e) {
            Log.error("failed to list element: "+tokens.sval);
         }
      }
      Log.trace("storage","getList returns "+ret); 
      return ret;
   }

   public long getListSize(OID cid) throws Exception {
      return getCollectionSize(cid);
   }

   public Object getListItem(OID cid, long index)
      throws Exception, IndexOutOfBoundsException
   {
      return getList(cid).get((int)index);
   }

   public boolean listContains(OID cid, Object value) 
      throws Exception
   {
      return getList(cid).contains(value);
   }

   protected void saveList(OID cid, List list) 
      throws IOException
   {
      Log.trace("storage","saveList("+cid+","+list+")");
      File file = new File(basedir,cid.toString());
      Log.trace("storage","file = "+file);
      PrintWriter writer = new PrintWriter(new FileWriter(file.toString(),false),true);
      for (int i=0; i<list.size(); i++) {
         String value = ValueConverter.objectToString(list.get(i));
         Log.trace("storage","  -> "+value);
         writer.println(Strings.slashify(value));
      }
   }

   public void addToList(OID cid, long position, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+position+","+value+")");
      List list = getList(cid);
      list.add((int)position,value);
      saveList(cid,list);
   }

   public void addToList(OID cid, Object value)
      throws Exception
   {
      Log.trace("storage","addToList("+cid+","+value+")");
      List list = getList(cid);
      list.add(value);
      saveList(cid,list);
   }

   public void setListItem(OID cid, long index, Object value)
      throws Exception
   {
      List list = getList(cid);
      list.set((int)index,value);
      saveList(cid,list);
   }

   public void removeFromList(OID cid, long position)
      throws Exception
   {
      List list = getList(cid);
      list.remove((int)position);
      saveList(cid,list);
   }

   public void removeFromList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      list.remove(value);
      saveList(cid,list);
   }

   public long getIndexInList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      return list.indexOf(value);
   }

   public long getLastIndexInList(OID cid, Object value)
      throws Exception
   {
      List list = getList(cid);
      return list.lastIndexOf(value);
   }


   // Set methods

   public void clearSet(OID cid) throws Exception
   {
      saveList(cid,new Vector());
   }

   public List getSet(OID oid, CollectionItem collection) 
      throws Exception 
   {
      return getSet(getCollectionID(oid,collection));
   }

   public List getSet(OID cid) 
      throws Exception 
   {
      return getList(cid);
   }

   public long getSetSize(OID cid) throws Exception {
      return getCollectionSize(cid);
   }

   public boolean addToSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","addToSet("+cid+","+value+")");
      List list = getList(cid);
      boolean ret = list.add(value);
      saveList(cid,list);
      return ret;
   }

   public boolean removeFromSet(OID cid, Object value) 
      throws Exception 
   {
      Log.trace("storage","addToSet("+cid+","+value+")");
      List list = getList(cid);
      boolean ret = list.remove(value);
      saveList(cid,list);
      return ret;
   }

   public boolean setContains(OID cid, Object value) 
      throws Exception
   {
      return getList(cid).contains(value);
   }

   // Map methods

   public Map getMap(OID oid, CollectionItem collection) 
      throws Exception
   {
      return getMap(getCollectionID(oid,collection));
   }

   public Map getMap(OID cid) 
      throws Exception
   {
      Log.trace("storage","getMap("+cid+")");
      File file = new File(basedir,cid.toString());
      StreamTokenizer tokens = getStreamTokenizer(file);
      Hashtable ret = new Hashtable();
      while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
         Object key = ValueConverter.stringToObject(Strings.unslashify(tokens.sval));
         tokens.nextToken();
         Object value = ValueConverter.stringToObject(Strings.unslashify(tokens.sval));
         Log.trace("storage",2,"  -> "+key+","+value);
         ret.put(key,value);
      }
      return ret;
   }

   public long getMapSize(OID cid) throws Exception {
      return getCollectionSize(cid);
   }

   protected void saveMap(OID cid, Map map) 
      throws IOException
   {
      Log.trace("storage","saveMap("+cid+","+map+")");
      File file = new File(basedir,cid.toString());
      Log.trace("storage","file = "+file);
      PrintWriter writer = new PrintWriter(new FileWriter(file.toString(),false),true);
      Set entrySet = map.entrySet();
      Iterator i = entrySet.iterator();
      while (i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         String value = ValueConverter.objectToString(entry.getValue());
         String key = ValueConverter.objectToString(entry.getKey());
         Log.trace("storage",2,"  -> "+key+","+value);
         writer.println(Strings.slashify(key)+" "+Strings.slashify(value));
      }
   }

   public void clearMap(OID cid) throws Exception
   {
      saveMap(cid,new Hashtable());
   }

   public Object putInMap(OID cid, Object key, Object value) 
      throws Exception
   {
      Map map = getMap(cid);
      Object ret = map.put(key,value);
      saveMap(cid,map);
      return ret;
   }

   public Object getFromMap(OID cid, Object key) 
      throws Exception
   {
      return getMap(cid).get(key);
   }

   public boolean mapContainsKey(OID cid, Object key)
      throws Exception
   {
      return getMap(cid).containsKey(key);
   }

   public boolean mapContainsValue(OID cid, Object value)
      throws Exception
   {
      return getMap(cid).containsValue(value);
   }

   public Object removeFromMap(OID cid, Object key)
      throws Exception
   {
      Log.trace("storage.map","removeFromMap("+cid+","+key+")");
      Map map = getMap(cid);
      Object ret = map.remove(key);
      saveMap(cid,map);
      return ret;
   }

   // others...

   public void removeField(OID oid, FieldItem field, Object value) 
      throws Exception
   {}

   public synchronized String newName(String className) throws Exception {
      semaphore.acquire();
      try {
         String name = nameGen.generateName(className);
         writeNameCounters();
         return name;
      } finally {
         semaphore.release();
      }
   }

   public OID getOIDFromName(String name) throws Exception {
      return (OID)names.get(name);
   }

   public String getNameFromOID(OID oid) throws Exception {
      return (String)oids.get(oid);
   }

   public void bindOIDToName(OID oid,String name) throws Exception
   {
      semaphore.acquire();
      try {
         names.put(name,oid);
         oids.put(oid,name);
         PrintWriter writer = 
            new PrintWriter(new FileWriter(oidsFile.toString(),true),true);
         writer.println(oid.toString()+" "+name);
      } finally {
         semaphore.release();
      }
   }

   public void deleteName(String name) throws Exception
   {}

   public String getClassID(OID oid) throws Exception {
      Log.trace("storage","getClassID("+oid+") -> "+(String)classes.get(oid));
      return (String)classes.get(oid);
   }

   public Collection getRootObjects() throws Exception {
      return oids.keySet();
   }

   public Collection getObjects(ClassItem cl) throws Exception {
      Log.trace("storage","getObjects("+cl+")");
      if (cl == null) {
         return classes.keySet();
      } else {
         Vector ret = new Vector();
         getObjects(cl,ret);
         return ret;
      }
   }

   /**
    * Gets all instances of a class and its subclasses.
    * @param cl the class
    * @param objects instances of the class are added to this collection
    */
   protected void getObjects(ClassItem cl, Vector objects) throws Exception {
      Log.trace("storage","getObjects("+cl+")");
      Set entries = classes.entrySet();
      Iterator i = entries.iterator();
      while(i.hasNext()) {
         Map.Entry entry = (Map.Entry)i.next();
         Log.trace("storage",2,"testing "+entry.getValue());
         if (cl==null || entry.getValue().equals(cl.getName())) {
            objects.add(entry.getKey());
         }
      }
      i = cl.getChildren().iterator();
      while(i.hasNext()) {
         ClassItem subclass = (ClassItem)i.next();
         getObjects(subclass,objects);
      }
   }


   public void startTransaction() {
   }

   public void commit() {
   }

   public void rollback() {
   }

}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***

