/*
  Copyright (C) 2002-2003 Laurent Martelli <laurent@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
 * ListWrapper.java
 *
 *
 * Created: Sat Oct 13 16:14:15 2001
 *
 * @author <a href="mailto: "Laurent Martelli</a>
 * @version
 */

package org.objectweb.jac.aspects.persistence;

import java.util.*;
import org.objectweb.jac.core.*;
import org.objectweb.jac.util.*;
import org.objectweb.jac.util.ExtBoolean;

/**
 * A wrapper for the list interface.
 */
public class ListWrapper extends CollectionWrapper {

   public ListWrapper(AspectComponent ac, boolean isLoaded) {
      super(ac,isLoaded);
   }

   protected void doLoad(Wrappee wrappee) throws Exception {
      OID oid = getOID(wrappee);
      List list = getStorage().getList(oid);
      for(int i=0; i<list.size(); i++) {
         try {
            list.set(i,normalizeOutput(list.get(i)));
         } catch (NoSuchOIDError e) {
            Log.error("ListWrapper.doLoad("+oid+
                      "): skipping object with unknown OID "+list.get(i));
            list.set(i,null);
         }
      }
      attrdef(ATTR_ADDED,"true");
      // Keep the number of dynamic invocations to the minimum
      Wrapping.invokeOrg(wrappee,"addAll",new Object[] { list } );
      attrdef(ATTR_ADDED,null);
   }

   public Object contains(Interaction interaction) throws Exception {
      if (isLoaded) {
         return proceed(interaction);
      } else {
         return ExtBoolean.valueOf(getStorage().listContains(getOID(interaction.wrappee),
                                                             normalizeInput(interaction.args[0])));
      }
   }

   public Object add(Interaction interaction) throws Exception {
      Log.trace("persistence",
                "adding "+interaction.args[0]+
                " to list "+getOID(interaction.wrappee));
      Object result = Boolean.TRUE;
      if (isLoaded)
         result = proceed(interaction);

      if (interaction.args.length==1) {
         getStorage().addToList(getOID(interaction.wrappee),
                                normalizeInput(interaction.args[0]));
      } else if (interaction.args.length==2) {
         getStorage().addToList(getOID(interaction.wrappee),
                                ((Integer)interaction.args[0]).longValue(),
                                normalizeInput(interaction.args[1]));
      }
      return result;
   }

   public Object get(Interaction interaction) throws Exception {
      if (isLoaded) {
         return proceed(interaction);
      } else {
         return normalizeOutput(
            getStorage().getListItem(getOID(interaction.wrappee),
                                     ((Integer)interaction.args[0]).longValue()));
      }
   }

   public Object remove(Interaction interaction) throws Exception {
      Log.trace("persistence.list", 
                "remove : "+interaction.wrappee+","+Arrays.asList(interaction.args));
      Object result = null;
      if (isLoaded) {
          result = proceed(interaction);
      } else {
         // WRONG!!!
         result = Boolean.TRUE;
      }
      Log.trace("persistence.list", "proceeded");
      if (interaction.args[0] instanceof Integer) {
         Log.trace("persistence.list", "remove position");
         getStorage().removeFromList(getOID(interaction.wrappee),
                                     ((Integer)interaction.args[0]).longValue());
      } else {
         Log.trace("persistence.list", "remove value");
         getStorage().removeFromList(getOID(interaction.wrappee),
                                     normalizeInput(interaction.args[0]));
      }
      Log.trace("persistence.list", "value removed");
      return result;
   }

   /**
    * Remove all instances from the collection
    */
   public Object clear(Interaction interaction) throws Exception {
      Log.trace("persistence.list", "clear");
      Object result = proceed(interaction);
      getStorage().clearList(getOID(interaction.wrappee));
      return result;
   }

   public Object set(Interaction interaction) throws Exception {
      if(isLoaded()) 
         proceed(interaction);
      getStorage().setListItem(getOID(interaction.wrappee),
                               ((Integer)interaction.args[0]).intValue(),
                               interaction.args[1]);
      return null;
   }

   public Object indexOf(Interaction interaction) throws Exception {
      if (isLoaded)
         return proceed(interaction);
      else 
         return new Integer(
            new Long(getStorage().getIndexInList(
               getOID(interaction.wrappee),
               normalizeInput(interaction.args[0]))).intValue() );
   }

   public Object lastIndexOf(Interaction interaction) throws Exception {
      if (isLoaded)
         return proceed(interaction);
      else 
         return new Integer( 
            new Long(getStorage().getLastIndexInList(
               getOID(interaction.wrappee),
               normalizeInput(interaction.args[0]))).intValue() );
   }

   protected long getCollectionSize(OID oid) throws Exception {
      return getStorage().getListSize(oid);
   }

   public Object iterator(Interaction interaction) {
      return new ListIterator(interaction.wrappee);
   }
}// ListWrapper

