/*
  Copyright (C) 2001-2003 Laurent Martelli <laurent@aopsys.com>
                          Renaud Pawlak <renaud@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

package org.objectweb.jac.aspects.persistence;

import gnu.regexp.RE;
import gnu.regexp.REException;
import org.objectweb.jac.aspects.naming.*;
import org.objectweb.jac.core.*;
import org.objectweb.jac.core.Wrappee;
import org.objectweb.jac.core.rtti.*;
import org.objectweb.jac.util.*;
import java.util.*;

/**
 * This AC defines a generic and configurable persistence aspect.<p>
 */

public class PersistenceAC extends AspectComponent 
   implements PersistenceConf {   

   /** The global identifier for a persistence root class in the
       RTTI. */
   public static final String ROOT = "root";

   /** The global identifier for a persistent class in the RTTI. */
   public static final String PERSISTENT = "persistent";

   public static final String VALUE_CONVERTER = "valueConverter";
   public static final String PRELOAD_FIELD = "preloadField";

   public static final String DISABLE_PRELOAD = "PersistenceAC.DISABLE_PRELOAD";
   public static final String NO_CACHE = "PersistenceAC.NO_CACHE";
   public static final String RESTORE = "PersistenceAC.RESTORE";

   // Object -> OID
   Hashtable oids = new Hashtable(); 
   // OID -> Object
   Hashtable objects = new Hashtable(); 

   private boolean connected = false;
   /**
    * The used storage. */
   protected Storage storage = null;
   /**
    * The loaded objects from the storage. */
   //   protected Hashtable objects = new Hashtable();

   ClassItem storageClass;
   String[] storageParameters;

   public void onExit() {
      if(storage!=null)
         storage.close();
   }

   //private HashSet statics = new HashSet();

    // ---- PersistenceConf interface


   public void setValueConverter(ClassItem cl, ClassItem converter) {
      try {
         cl.setAttribute(VALUE_CONVERTER, converter.newInstance());
      } catch (Exception e) {
         throw new RuntimeException(e.getMessage());
      }
   }

   /**
    * This method is a callback for the timer that defines the max
    * idle time.
    *
    * <p>For all the collections that have a max idle time, it checks
    * that this time is not reached. If it is reached, it unloads the
    * collection.
    *
    * @see #defineMaxIdleCheckPeriod(long)
    * @see #maxIdle(CollectionItem,long) */

   public void checkUnload() {
      
      Iterator it = collectionIdles.entrySet().iterator();
      while(it.hasNext()) {
         Map.Entry entry = (Map.Entry)it.next();

         CollectionItem coll = (CollectionItem)entry.getKey();
         long idle = ((Long)entry.getValue()).longValue();

         Log.trace("persistence.cache","checking collection "+coll+" (idle="+idle+")");
         ClassItem cl = coll.getClassItem();
         Object[] objects = ObjectRepository.getMemoryObjects(cl);
         for(int i=0; i<objects.length; i++) {
            Wrappee wrappee=(Wrappee)coll.get(objects[i]);
            Log.trace("persistence.cache","checking wrappee "+
                      NameRepository.get().getName(objects[i])+
                      "."+wrappee);
            Date useDate = (Date)Wrapping.invokeRoleMethod(
               wrappee,CollectionWrapper.class,"getUseDate",new Object[0]);
            Date now = new Date();
            long lastUsed = now.getTime() - useDate.getTime();
            Log.trace("persistence.cache","last used="+
                      lastUsed+" ms ago");
            if (lastUsed>idle) {
               Log.trace("persistence.cache","unloading collection "+
                         NameRepository.get().getName(objects[i])+
                         "."+coll);
               Wrapping.invokeRoleMethod(wrappee,CollectionWrapper.class,
                                         "unload",new Object[0]);
            }
         }
      }
   }

   long checkPeriod = -1;
   
   public void defineMaxIdleCheckPeriod(long period) {
      checkPeriod = period;
   }

   HashMap collectionIdles = new HashMap();

   public void maxIdle(CollectionItem collection,long maxIdle) {
      collectionIdles.put(collection,new Long(maxIdle));
   }

   public void whenConfigured() {
      if (checkPeriod==-1) {
         checkPeriod = 200000;
      }
      defineTimer(checkPeriod,
                  ClassRepository.get()
                  .getClass(PersistenceAC.class)
                  .getMethod("checkUnload"),
                  new Object[] {});
   }

   // Pointcut for roots objects
   MethodPointcut rootsPointcut;

   public void configureStorage(ClassItem storageClass, 
                                String[] storageParameters) {
      this.storageClass = storageClass;
      this.storageParameters = storageParameters;
      String nameExpr = "";
      Iterator roots = RttiAC.getRoots().iterator();
      while(roots.hasNext()) {
         String nextRoot = (String)roots.next();
         nameExpr = nameExpr+nextRoot;
         if (roots.hasNext()) {
            nameExpr = nameExpr+" || ";
         }
      }
      rootsPointcut = (MethodPointcut)pointcut(
         nameExpr,"ALL","CONSTRUCTORS",
         PersistenceWrapper.class.getName(),"handleStatic",null,false);
      Log.trace("persistence.statics","storage configured, roots pointcut='"+
                nameExpr+"'");
   }

   /** MethodPointcuts for static objects */
   Vector staticPointcuts = new Vector();

   public void registerStatics(String classExpr,String nameExpr) 
   {
      /* Wrap constructors with PersistenceWrapper.handleStatic() */
      Log.trace("persistence","registerStatics "+classExpr+" "+nameExpr);
      staticPointcuts.add(
         pointcut(nameExpr,classExpr,"CONSTRUCTORS",
                  PersistenceWrapper.class.getName(),"handleStatic",null,false)
      );
   }

   /** Pointcuts for persistent objects */
   Vector persistentPointcuts = new Vector();

   public void makePersistent(String classExpr,String nameExpr) 
   {
      // Wrap modifiers, collection accessors, reference accessors
      // (but not constructors) with PersistenceWrapper.applyPersistence()
      persistentPointcuts.add(
         pointcut(nameExpr,classExpr,
                  "MODIFIERS({!transient}) || COLACCESSORS({!transient}) || REFACCESSORS({!transient}) && !CONSTRUCTORS",
                  PersistenceWrapper.class.getName(),"applyPersistence",null,true)
      );
   }

   /**
    * Tells wether a wrappee is persistent or not.
    * @param wrappee the wrappee
    * @see #makePersistent(String,String)
    */
   boolean isPersistent(Wrappee wrappee) {
      Iterator it = persistentPointcuts.iterator();
      ClassItem cli = ClassRepository.get().getClass(wrappee);
      while (it.hasNext()) {
         MethodPointcut pointcut = (MethodPointcut)it.next();
         if (pointcut.isClassMatching(wrappee,cli))
            return true;
      }
      return false;
   }


   /**
    * Tells wether a wrappee is static or not.
    * @param wrappee the wrapee
    * @param objName name of the wrappee
    * @see #makePersistent(String,String)
    */
   boolean isStatic(Wrappee wrappee, String objName) {
      ClassItem cli = ClassRepository.get().getClass(wrappee);
      // check the roots
      if(rootsPointcut.isClassMatching(wrappee,cli) &&
         rootsPointcut.isNameMatching(wrappee,objName)) {
         return true;
      }
      // check the statics
      Iterator it = staticPointcuts.iterator();
      while (it.hasNext()) {
         MethodPointcut pointcut = (MethodPointcut)it.next();
         if (pointcut.isClassMatching(wrappee,cli) &&
             pointcut.isNameMatching(wrappee,objName))
            return true;
      }
      return false;
   }

   // ---- end of PersistenceConf

   /**
    * Returns the storage.
    *
    * @return the storage
    */
   public Storage getStorage() {
      if (storage == null) {
         if (storageClass==null) {
            throw new RuntimeException("Persistence: storage is not configured");
         }
         Log.trace("persistence",getClass().getName()+".connectStorage("+
                   storageClass+","+
                   (storageParameters!=null?Arrays.asList(storageParameters).toString():"null")+
                   ")");
         try {
            Iterator i = storageClass.getConstructors().iterator();
            ConstructorItem constructor = null;
            while (i.hasNext()) {
               ConstructorItem c = (ConstructorItem)i.next();
               if (c.getParameterTypes().length == storageParameters.length) {
                  constructor = c;
                  break;
               }
            }
            if (constructor != null) {
               storage = (Storage)constructor.newInstance(storageParameters);
            } else {
               throw new Exception(
                  "Could not find a constructor for class "+storageClass+
                  " with parameters "+Arrays.asList(storageParameters));
            }
         } catch (Exception e) {
            Log.error("failed to connect to storage : "+e);
            e.printStackTrace();
         }
      }
      return storage;
   }

   /**
    * The persistence aspect checks whether an object was in the
    * storage when a <code>NameRepository.getObject</code> call
    * failed.
    * @see BaseProgramListener#whenObjectMiss(String)
    */
   public void whenObjectMiss(String name) {
      Log.trace("persistence.naming","whenObjectMiss "+name);
      try {
         OID oid = getStorage().getOIDFromName(name);
         if (oid!=null) {
            Log.trace("persistence.naming","found name "+name+" -> "+oid);
            attrdef(BaseProgramListener.FOUND_OBJECT,getObject(oid,null));
         }
      } catch (Exception e) {
         e.printStackTrace();
      }
   }

   /**
    * Delegates naming to the storage
    */
   public String whenNameObject(Object object, String name) {
      Log.trace("persistence.naming","whenNameObject "+object+" <- "+name);
      Log.stack("persistence.naming."+object.getClass().getName());
      if (!(object instanceof Wrappee))
         return name;
      Wrappee wrappee = (Wrappee)object;
      if (isPersistent(wrappee) && !isStatic(wrappee,name)) {
         try {
            name = getStorage().newName(object.getClass().getName());
         } catch (Exception e) {
            Log.error("Failed to name object "+object+": "+e);
            e.printStackTrace();
         }
         Log.trace("persistence.naming","   -> "+name);
      }
      return name;
   }

   /**
    * Add an object in the list of persistent objects.
    *
    * @param oid the OID of the object
    * @param object the object
    */
   protected void registerObject(OID oid, Object object) {
      Log.trace("persistence","registerObject("+oid+","+object.getClass()+")");
      //      Log.trace("persistence",2,"objects = "+objects);
      //      new Exception().printStackTrace();
      if (objects.containsKey(oid))
         throw new Error("Persistence: an object is already registered with this oid: "+oid);
      objects.put(oid,object);
      oids.put(object,oid);
      Log.trace("persistence",2,"object "+oid+" added");
   }

   /**
    * Returns a reference to an object with a given OID.
    * 
    * <p>Loads the object from a storage if necessary, or returns a
    * cached object.</p>
    *
    * @param oid OID of the object 
    * @param newObject use this object instead of instanciating a new one
    */
   synchronized Object getObject(OID oid, Object newObject) {
      Object result = objects.get(oid);
      try {
         if (result != null) {
            Log.trace("persistence",2,
                      "Object "+oid+" found in cache -> "+result);
            return result;
         } else {
            Log.trace("persistence",
                      this+".Object "+oid+
                      " NOT found in cache; Loading from storage\n");
            String lClassID = getStorage().getClassID(oid);
            if (lClassID==null)
               Log.error("persistence: getClassID("+oid+") -> NULL");
            ClassItem lClass = ClassRepository.get().getClass(lClassID);
            Log.trace("persistence","Class = "+lClass.getName());
            if (newObject==null) {
               Naming.setName(getStorage().getNameFromOID(oid));
               Collaboration collab = Collaboration.get();
               collab.addAttribute(RESTORE,Boolean.TRUE);
               try {
                  newObject = lClass.newInstance();
               } finally {
                  collab.removeAttribute(RESTORE);
               }
            }
            Wrappee wrappee = (Wrappee)newObject;
            //PersistenceWrapper wrapper = wrap(wrappee,oid);
            registerObject(oid,wrappee);
            // load the new object's fields
            Wrapping.invokeRoleMethod(wrappee,
                                      PersistenceWrapper.class,
                                      "loadAllFields",
                                      new Object[] {});
            // wrap its collections
            Wrapping.invokeRoleMethod(wrappee,
                                      PersistenceWrapper.class,
                                      "wrapCollections",
                                      new Object[] {Boolean.FALSE});
            Log.trace("persistence","New object "+oid+" : "+newObject);
            Log.trace("persistence","Object loaded");
            result=newObject;
         }
      } catch(Exception e) {
         e.printStackTrace();
      }
      return result;
   }

   public OID getOID(Wrappee wrappee) {
      return (OID)oids.get(wrappee);
   }

   /**
    * Gets a persistent object in the storage from its OID.
    *
    * <p>By default, this method loads all the fields that are
    * primitive to initialize the object state. Collections and
    * references are initialized later on when they are accessed by
    * some methods.<p>
    *
    * @param oid the OID of the seeked object
    * @return the object
    *
    * @see PersistenceWrapper#initAllFields(Wrappee) */
 
   /*
   public synchronized Object getObject(OID oid) throws Exception {

      Object result = objects.get( oid );

      if ( result != null ) {
         Log.trace("persistence","Object "+oid+" found in cache");
         return result;
      } else {
         Log.trace("persistence",
                   this+".Object "+oid+" NOT found in cache; Loading from storage\n");
         String lClassID = getStorage().getClassID( oid );
         if (lClassID==null)
            Log.error("persistence: getClassID("+oid+") -> NULL");
         ClassItem lClass = ClassRepository.get().getClass(lClassID);
         Log.trace("persistence","Class = "+lClass.getName());
         attrdef("forceName", getStorage().getNameFromOID(oid));
         Object newObject = lClass.newInstance();
         
         Wrappee wrappee = (Wrappee)newObject;
         //PersistenceWrapper wrapper = wrap(wrappee,oid);
         Log.trace("persistence",
                   "New object " + oid + " : " +newObject.getClass().getName() );
         Log.trace("persistence","Object loaded");
         return newObject;
      }
   }
   */

   /**
    * This method allows the deserialization of the OID of a
    * persistent object.<p>
    *
    * @param orgObject the JAC object structure that is beeing
    * deserialized
    * @see #whenSerialized(SerializedJacObject) */

   public void whenDeserialized ( SerializedJacObject orgObject ) {
      /*      
      Log.trace("persistence","PersistenceAC.whenDeserialize");
      OID oid = (OID) orgObject.getACInfos( "persistence" );
      if ( oid != null ) {
         Wrappee finalObject = (Wrappee)attr("finalObject");
         PersistenceWrapper pw = wrap( finalObject, oid, false );
         //pw.setOid( oid );
      }
      */
   }

   /**
    * This method add the OID info to the serialized JAC object when a
    * serialization is requested by another aspect.<p>
    *
    * This adding allows, for instance, the OID to be transmitted to
    * remote containers.<p>
    *
    * @param finalObject the object that is being serialized */

   public void whenSerialized ( SerializedJacObject finalObject ) {
      /*  
      Log.trace("persistence","PersistenceAC.whenSerialize");
      Wrappee orgObject = (Wrappee)attr("orgObject");
      if ( orgObject.isExtendedBy( PersistenceWrapper.class ) ) {
         finalObject.setACInfos(
            "persistence", orgObject.invokeRoleMethod( "getOid", new Object[0] ) );
      }
      */
   }

   /**
    * Load objects from the storage when required.
    */
   public void whenGetObjects(Collection objects, ClassItem cl)
   {
      Log.trace("getObjects","PersistenceAC.whenGetObjects "+cl);
      if (cl == null) 
         return;
      try {
         Log.trace("getObjects","PersistenceAC.whenGetObjects "+cl);
         Collection oids = getStorage().getObjects(cl);
         Iterator i = oids.iterator();
         while (i.hasNext()) {
            OID oid = (OID)i.next();
            Object object = getObject(oid,null);
            if (!objects.contains(object))
               objects.add(object);
         }
      } catch (Exception e) {
         e.printStackTrace();
      }
   }

   public boolean beforeRunningWrapper( Wrapper wrapper, 
                                        String wrappingMethod ) {
      if (attr("Persistence.disabled") != null) {
         return false;
      } else {
         return true;
      }
   }

   public void whenDeleted(Wrappee object) {
      if(storage!=null) {
         try {
            OID oid = getOID(object);
            if (oid!=null)
               storage.deleteObject(oid);
            whenFree(object);
         } catch(Exception e) {
            e.printStackTrace();
         }
      }
   }

   public void whenFree(Wrappee object) {
      if (storage!=null) {
         try {
            OID oid = getOID(object);
            if (oid!=null)
               oids.remove(oid);
            if (objects.contains(object))
               objects.remove(object);
         } catch(Exception e) {
            e.printStackTrace();
         }
      }
   }

   public void preloadField(FieldItem field) {
      field.setAttribute(PRELOAD_FIELD, Boolean.TRUE);
   }

   public void preloadAllFields(ClassItem cl) {
      FieldItem[] fields = cl.getFields();
      if (fields!=null) {
         for(int i=0; i<fields.length; i++) {
            fields[i].setAttribute(PRELOAD_FIELD, Boolean.TRUE);
         }
      }
   }

   public static boolean isFieldPreloaded(FieldItem field) {
      Boolean value = (Boolean)field.getAttribute(PRELOAD_FIELD);      
      return value!=null && value.booleanValue();
   }

   public void disableCache(CollectionItem collection) {
      collection.setAttribute(NO_CACHE,"true");
   }

   public String[] getDefaultConfigs() {
      return new String[] {
         "org/objectweb/jac/aspects/persistence/persistence.acc",
         "org/objectweb/jac/aspects/user/persistence.acc"};
   }

   public static class NoSuchWrapperException extends RuntimeException {}

}

// Local Variables: ***
// c-basic-offset:3 ***
// End: ***
